"""clovers 0.4"""

__version__ = "0.4"
__description__ = "高度自定义的聊天平台 Python 异步机器人指令-响应插件框架"

from .clovers import Leaf, Client, LeafClient
from .core import Result, Event, BaseHandle, Handle, TempHandle, Plugin, Adapter
